import os, requests, pkg_resources, re, socket
from shutil import copy, copytree
from pathlib import Path
from packaging import version
from uniquebible import config


# go to resource directory
thisFile = os.path.realpath(__file__)
config.packageDir = os.path.dirname(thisFile)

ubahome = os.path.expanduser(os.path.join("~", "UniqueBible"))

# restor previous backup
configFile = os.path.join(config.packageDir, "config.py")
backupFile = os.path.join(ubahome, "config.py")
if os.path.isfile(backupFile) and not hasattr(config, "mainText"):
    print(f"Configuration backup found: {backupFile}")
    try:
        print("Loading backup ...")
        with open(backupFile, "r", encoding="utf-8") as fileObj:
            configs = fileObj.read()
        # load backup configs
        configs = "from uniquebible import config\n" + re.sub("^([A-Za-z0-9])", r"config.\1", configs, flags=re.M)
        exec(configs, globals())
        # copy backup configs
        copy(backupFile, configFile)
        print("Previous configurations restored!")
    except:
        try:
            os.rename(backupFile, f"{backupFile}_failed")
        except:
            pass
        print("Failed to restore previous configurations!")

# set up folders for storing user content in ~/UniqueBible
if not os.path.isdir(ubahome):
    Path(ubahome).mkdir(parents=True, exist_ok=True)
for i in ("audio", "htmlResources", "import", "macros", "marvelData", "music", "notes", "temp", "terminal_history", "terminal_mode", "thirdParty", "video", "webstorage", "workspace"):
    sourceFolder = os.path.join(config.packageDir, i)
    targetFolder = os.path.join(ubahome, i)
    if not os.path.isdir(targetFolder):
        print(f"Setting up user directory '{i}' ...")
        copytree(sourceFolder, targetFolder, dirs_exist_ok=True)
    #copytree(i, os.path.join(ubahome, i), dirs_exist_ok=True)

# set up map images
import uniquebible.htmlResources.images.exlbl
import uniquebible.htmlResources.images.exlbl_large
import uniquebible.htmlResources.images.exlbl_largeHD_1
import uniquebible.htmlResources.images.exlbl_largeHD_2
import uniquebible.htmlResources.images.exlbl_largeHD_3

# Add folders below for all new folders created after version 0.1.17
# ...

# change directory to user directory
if os.getcwd() != ubahome:
    os.chdir(ubahome)

# user plugins; create folders for users to place their own plugins
# TODO: user plugins not working for now; will implement later
for i in ("chatGPT", "config", "context", "event", "language", "layout", "menu", "shutdown", "startup", "terminal", "text_editor"):
    Path(os.path.join(ubahome, "plugins", i)).mkdir(parents=True, exist_ok=True)


def getPackageInstalledVersion(package):
    try:
        installed_version = pkg_resources.get_distribution(package).version
        return version.parse(installed_version)
    except pkg_resources.DistributionNotFound:
        return None

def getPackageLatestVersion(package):
    try:
        response = requests.get(f"https://pypi.org/pypi/{package}/json", timeout=10)
        latest_version = response.json()['info']['version']
        return version.parse(latest_version)
    except:
        return None

def isServerAlive(ip, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.settimeout(2)  # Timeout in case of server not responding
    try:
        sock.connect((ip, port))
        sock.close()
        return True
    except socket.error:
        return False

if hasattr(config, "checkVersionOnStartup") and config.checkVersionOnStartup:

    if isServerAlive("8.8.8.8", 53):

        thisPackage = "uniquebible"

        content = []

        content.append(f"# Checked '{thisPackage}' version ...")

        config.version = installed_version = getPackageInstalledVersion(thisPackage)
        if installed_version is not None:
            content.append(f"Installed version: {installed_version}")
            latest_version = getPackageLatestVersion(thisPackage)
            if latest_version is not None:
                content.append(f"Latest version: {latest_version}")
                if latest_version > installed_version:
                    content.append("Run `pip install --upgrade uniquebible` to upgrade!")
                    print("\n".join(content))

        config.internetConnectivity = True
    else:
        config.internetConnectivity = False
        config.version = "0.0.0"

else:
    config.version = "0.0.0"

# AI Features

from openai import OpenAI
from mistralai import Mistral
from groq import Groq
from typing import Optional
import unicodedata

def is_CJK(self, text):
    for char in text:
        if 'CJK' in unicodedata.name(char):
            return True
    return False

def isLLMReady():
    if config.llm_backend == "openai" and config.openaiApi_key:
        return True
    elif config.llm_backend == "mistral" and config.mistralApi_key:
        return True
    elif config.llm_backend == "groq" and config.groqApi_key:
        return True
    return False

def getGroqApi_key() -> str:
    '''
    support multiple grop api keys
    User can manually edit config to change the value of config.groqApi_key to a list of multiple api keys instead of a string of a single api key
    '''
    if config.groqApi_key:
        if isinstance(config.groqApi_key, str):
            return config.groqApi_key
        elif isinstance(config.groqApi_key, list):
            if len(config.groqApi_key) > 1:
                # rotate multiple api keys
                config.groqApi_key = config.groqApi_key[1:] + [config.groqApi_key[0]]
            return config.groqApi_key[0]
        else:
            return ""
    else:
        return ""

def getMistralApi_key() -> str:
    '''
    support multiple mistral api keys
    User can manually edit config to change the value of config.mistralApi_key to a list of multiple api keys instead of a string of a single api key
    '''
    if config.mistralApi_key:
        if isinstance(config.mistralApi_key, str):
            return config.mistralApi_key
        elif isinstance(config.mistralApi_key, list):
            if len(config.mistralApi_key) > 1:
                # rotate multiple api keys
                config.mistralApi_key = config.mistralApi_key[1:] + [config.mistralApi_key[0]]
            return config.mistralApi_key[0]
        else:
            return ""
    else:
        return ""

def getChatResponse(backend, chatMessages) -> Optional[str]:
    if not isLLMReady():
        return None
    try:
        if backend == "mistral":
            completion = Mistral(api_key=getMistralApi_key()).chat.complete(
                model=config.mistralApi_chat_model,
                messages=chatMessages,
                n=1,
                temperature=config.mistralApi_llmTemperature,
                max_tokens=config.mistralApi_chat_model_max_tokens,
                stream=False,
            )
        elif backend == "openai":
            os.environ["OPENAI_API_KEY"] = config.openaiApi_key
            completion = OpenAI().chat.completions.create(
                model=config.openaiApi_chat_model,
                messages=chatMessages,
                n=1,
                temperature=config.openaiApi_llmTemperature,
                max_tokens=config.openaiApi_chat_model_max_tokens,
                stream=False,
            )
        else:
            completion = Groq(api_key=getGroqApi_key()).chat.completions.create(
                model=config.groqApi_chat_model,
                messages=chatMessages,
                n=1,
                temperature=config.groqApi_llmTemperature,
                max_tokens=config.groqApi_chat_model_max_tokens,
                stream=False,
            )
        textOutput = completion.choices[0].message.content
    except:
        textOutput = "Failed to connect! Please try again later."
    return textOutput

def getAiFeatureDisclaimer() -> str:
    if config.displayLanguage == "zh_HANT":
        disclaimer = """<p><b>免責聲明：</b> 本網站上由 AI 提供支援的聖經功能旨在提供有關聖經的有用信息和見解。然而，它不能代替個人對經文的學習和反思。聖經本身仍然是基督徒真理和權威的最終來源。請僅將此工具提供的資訊用作參考，並始終查閱聖經以獲得有關您問題的明確答案。</p>"""
    elif config.displayLanguage == "zh_HANS":
        disclaimer = """<p><b>免责声明：</b> 本网站上由 AI 提供支持的圣经功能旨在提供有关圣经的有用信息和见解。然而，它不能代替个人对经文的学习和反思。圣经本身仍然是基督徒真理和权威的最终来源。请仅将此工具提供的信息用作参考，并始终查阅圣经以获得有关您问题的明确答案。</p>"""
    else:
        disclaimer = """<p><b>Disclaimer:</b> The AI-powered Bible feature on this website is intended to provide helpful information and insights about the Bible. However, it is not a substitute for personal study and reflection on the scriptures. The Bible itself remains the ultimate source of truth and authority for Christians. Please use the information provided by this tool for reference only, and always consult the Bible for definitive answers to your questions.</p>"""
    return disclaimer
